from referencing import Registry, Resource
from jsonschema import Draft202012Validator
import json


def get_validator():
    template_resource = Resource.from_contents(template_schema)
    chain_resource = Resource.from_contents(chain_schema)
    llm_resource = Resource.from_contents(llm_schema)
    prompt_resource = Resource.from_contents(prompt_schema)
    memory_resource = Resource.from_contents(memory_schema)

    registry = chain_resource @ (
        llm_resource
        @ (prompt_resource @ (memory_resource @ (template_resource @ Registry())))
    )

    validator = Draft202012Validator(template_schema, registry=registry)

    return validator


def validate_chain(json):
    """Validate chain template json against json schema.

    Args:
        json (string): The json to validate

    Raises:
        ValidationError: The schema validation error
    """
    validator = get_validator()
    validator.validate(json)


def validate_file(path):
    """Validate chain template json against json schema.

    Args:
        path (string): The path to the file containing json

    Raises:
        ValidationError: The schema validation error
    """
    with open(path) as f:
        cfg = json.load(f)
    validate_chain(cfg)


template_schema = {
    "$id": "https://example.com/template.schema.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "LangChain Template",
    "description": "A langchain template",
    "type": "object",
    "properties": {
        "name": {"type": "string", "description": "The name of the langchain template"},
        "description": {
            "type": "string",
            "description": "Description of the langchain template",
        },
        "template_version": {
            "type": "string",
            "description": "Version of the langchain template interpreter.",
        },
        "chain": {"$ref": "https://example.com/chain.schema.json"},
    },
    "additionalProperties": False,
    "required": ["name", "template_version", "chain"],
}

chain_schema = {
    "$id": "https://example.com/chain.schema.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "LangChain Chain",
    "description": "A langchain chain",
    "type": "object",
    "properties": {
        "type": {"type": "string", "description": "The type of chain to be used"},
        "llm": {
            "$ref": "https://example.com/llm.schema.json",
            "description": "Specification of the llm to be used",
        },
        "chains": {
            "type": "array",
            "description": "A list of chains to be executed in the sequential chain.",
            "items": {"$ref": "https://example.com/chain.schema.json"},
        },
        "prompt": {"$ref": "https://example.com/prompt.schema.json"},
        "memory": {"$ref": "https://example.com/memory.schema.json"},
        "input_variables": {
            "type": "array",
            "description": "List of input variables to the chain",
            "items": {"type": "string"},
        },
        "output_variables": {
            "type": "array",
            "description": "List of output variables for the chain",
            "items": {"type": "string"},
        },
        "output_key": {
            "type": "string",
            "description": "The key for the output of this specific chain.",
        },
    },
    "additionalProperties": False,
    "required": ["type"],
}

llm_schema = {
    "$id": "https://example.com/llm.schema.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "LangChain LLM",
    "description": "A langchain llm",
    "type": "object",
    "properties": {
        "type": {
            "type": "string",
            "description": "The type of llm to be used",
        },
        "args": {
            "type": "object",
            "description": "The arguments of the llm to be passed in",
        },
    },
    "additionalProperties": False,
    "required": ["type"],
}

prompt_schema = {
    "$id": "https://example.com/prompt.schema.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "LangChain Prompt Template",
    "description": "A langchain prompt template",
    "type": "object",
    "properties": {
        "template": {
            "type": "string",
            "description": "The template string for the prompt template.",
        },
        "input_variables": {
            "type": "array",
            "description": "A list of the input variables for the template",
            "items": {"type": "string"},
        },
    },
    "additionalProperties": False,
    "required": ["template"],
}

memory_schema = {
    "$id": "https://example.com/memory.schema.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "LangChain Memory",
    "description": "A langchain memory object",
    "type": "object",
    "properties": {
        "type": {
            "type": "string",
            "description": "The type of memory to use.",
        },
        "memories": {
            "type": "object",
            "description": "A kv store of the memories for simple memory.",
            "additionalProperties": {"type": "string"},
        },
        "context": {
            "type": "array",
            "description": "List of conversation inputs and outputs.",
            "items": {"type": "object"},
        },
        "history": {
            "type": "string",
            "description": "String of previous conversation.",
        },
    },
    "additionalProperties": False,
    "required": ["type"],
}



from langchain.chains import (
    LLMChain,
    SimpleSequentialChain,
    SequentialChain,
    ConversationChain,
)
from langchain import PromptTemplate, OpenAI, LLMChain
from langchain.memory import SimpleMemory, ConversationBufferMemory
from langchain.llms import type_to_cls_dict
from langchain_interpreter import validate_chain
import json


def chain_from_file(filename, **kwargs):
    """Generate a LangChain Chain from a json file.

    Args:
        filename (str): the path to the file containing json template.

    Returns:
        chain: the chain specified by the json template.

    Raises:
        ValidationError:
    """

    with open(filename) as f:
        cfg = json.load(f)
    validate_chain(cfg)
    return get_chain(cfg["chain"], **kwargs)


def chain_from_str(s, **kwargs):
    """Generate a LangChain Chain from a json string.

    Args:
        s (str): the json string to turn into a chain

    Returns:
        chain: the chain specified by the json template.
    """

    cfg = json.loads(s)
    validate_chain(cfg)
    return get_chain(cfg["chain"], **kwargs)


def get_chain(cfg, **kwargs):
    if cfg["type"] == "LLMChain":
        return get_llm_chain(cfg, **kwargs)
    if cfg["type"] == "SimpleSequentialChain":
        return get_simple_sequential_chain(cfg, **kwargs)
    if cfg["type"] == "SequentialChain":
        return get_sequential_chain(cfg, **kwargs)
    if cfg["type"] == "ConversationChain":
        return get_conversation_chain(cfg, **kwargs)


def get_conversation_chain(cfg, **kwargs):
    llm = get_llm(cfg["llm"], **kwargs)
    memory = try_memory(cfg, **kwargs)

    conv_chain = ConversationChain(llm=llm, memory=memory)
    return conv_chain


def get_llm(cfg, **kwargs):
    # TODO: restrict types you can use
    llm_type = type_to_cls_dict[cfg["type"]]
    llm = llm_type(**cfg["args"], openai_api_key=kwargs.get("openai_api_key"))
    return llm


def get_llm_chain(cfg, **kwargs):
    llm = get_llm(cfg["llm"], **kwargs)
    prompt = get_prompt(cfg["prompt"])
    output_key = cfg.get("output_key", "text")
    memory = try_memory(cfg, **kwargs)
    llm_chain = LLMChain(llm=llm, prompt=prompt, output_key=output_key, memory=memory)
    return llm_chain


def get_simple_sequential_chain(cfg, **kwargs):
    chains = [get_chain(chain, **kwargs) for chain in cfg["chains"]]
    memory = try_memory(cfg, **kwargs)
    ss_chain = SimpleSequentialChain(chains=chains, verbose=True, memory=memory)
    return ss_chain


def get_prompt(cfg):
    prompt = PromptTemplate(**cfg)
    return prompt


def get_sequential_chain(cfg, **kwargs):
    chains = [get_chain(chain, **kwargs) for chain in cfg["chains"]]
    input_variables = cfg["input_variables"]
    output_variables = cfg["output_variables"]
    memory = try_memory(cfg)
    seq_chain = SequentialChain(
        chains=chains,
        input_variables=input_variables,
        output_variables=output_variables,
        verbose=True,
        memory=memory,
    )
    return seq_chain


def try_memory(cfg, **kwargs):
    if "memory" in cfg:
        return get_memory(cfg["memory"], **kwargs)
    return None


def get_memory(cfg, **kwargs):
    if cfg["type"] == "SimpleMemory":
        return get_simple_memory(cfg, **kwargs)
    if cfg["type"] == "ConversationBufferMemory":
        return get_conversation_buffer_memory(cfg, **kwargs)


def get_simple_memory(cfg, **kwargs):
    memories = cfg["memories"]
    simple_memory = SimpleMemory(memories=memories)
    return simple_memory


def get_conversation_buffer_memory(cfg, **kwargs):
    conv_buf_memory = ConversationBufferMemory()

    if "context" in cfg:
        context = cfg["context"]
        if len(context) % 2 != 0:
            raise ValueError("the context doesn't have an even number of elements!")
    elif "history" in cfg:
        history = cfg["history"]
        context = parse_history(history)
    else:
        context = []

    if "history" in kwargs:
        history = kwargs.get("history")
        context += parse_history(history)

    # Save context
    for pos in range(0, len(context), 2):
        conv_buf_memory.save_context(context[pos], context[pos + 1])

    return conv_buf_memory


def parse_history(history: str):
    if len(history) == 0:
        return []

    lines = history.split("\n")
    ctx = []

    for l in lines:
        if l.startswith("Human: "):
            ctx.append({"input": l.lstrip("Human: ")})
        elif l.startswith("AI: "):
            ctx.append({"output": l.lstrip("AI: ")})
        else:
            raise ValueError("each line in history must be from AI or human")

    return ctx
